package be.iminds.ilabt.jfed.experimenter_gui.connectivity_tester;

import be.iminds.ilabt.jfed.connectivity_tester.ConnectivityTest;
import be.iminds.ilabt.jfed.experimenter_gui.util.ui.StageUtils;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.beans.binding.Bindings;
import javafx.collections.MapChangeListener;
import javafx.fxml.FXML;
import javafx.scene.Scene;
import javafx.scene.control.Accordion;
import javafx.scene.control.Button;
import javafx.scene.control.ScrollPane;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.HBox;
import javafx.stage.FileChooser;
import javafx.stage.Stage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import java.io.File;
import java.io.IOException;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.HashMap;
import java.util.Map;

/**
 * User: twalcari
 * Date: 1/3/14
 * Time: 9:49 AM
 */
public class ConnectivityTesterGUI extends BorderPane {
    private static final Logger LOG = LoggerFactory.getLogger(ConnectivityTesterGUI.class);
    private static final String TESTRESULT_EXTENSION = ".txt";
    private static final DateTimeFormatter FILE_DATE_TIME_FORMATTER = DateTimeFormatter.ofPattern("yyyy-MM-dd-HH-mm").withZone(ZoneId.systemDefault());

    private final ConnectivityTester connectivityTester;
    private final JFedGuiPreferences jFedPreferences;

    private Stage stage;
    @FXML
    private ScrollPane testsScrollPane;
    @FXML
    private Accordion testsAccordion;
    @FXML
    private Button saveButton;
    @FXML
    private Button restartButton;

    @FXML
    private HBox initializationBox;


    private Map<ConnectivityTest, ConnectivityTestPane> testPanes = new HashMap<>();
    private boolean testsStarted = false;

    @Inject
    public ConnectivityTesterGUI(ConnectivityTester connectivityTester, JFedGuiPreferences jFedPreferences) {
        this.connectivityTester = connectivityTester;
        this.jFedPreferences = jFedPreferences;

        FXMLUtil.injectFXML(this);

        restartButton.disableProperty().bind(connectivityTester.testsFinishedProperty().not());
        saveButton.disableProperty().bind(connectivityTester.testsFinishedProperty().not());

        testsAccordion.prefWidthProperty().bind(Bindings.subtract(testsScrollPane.widthProperty(), 15));

        connectivityTester.getTestTasks().addListener((MapChangeListener<ConnectivityTest, ConnectivityTestTask>) change -> {
            if (change.wasAdded()) {
                ConnectivityTestPane pane = new ConnectivityTestPane(change.getValueAdded());
                pane.prefWidthProperty().bind(testsAccordion.widthProperty());
                testsAccordion.getPanes().add(pane);

                testPanes.put(change.getKey(), pane);
            }

            if (change.wasRemoved()) {
                ConnectivityTestPane pane = testPanes.get(change.getKey());
                testsAccordion.getPanes().remove(pane);
            }
        });

        initializationBox.visibleProperty().bind(Bindings.isEmpty(testsAccordion.getPanes()));

    }

    public synchronized void showDialog() {
        if (stage == null) {
            stage = new Stage();
            stage.setTitle("jFed Connectivity Tester");
            StageUtils.setJFedStageIcons(stage);
            stage.setScene(new Scene(this));
        }

        stage.show();

        connectivityTester.startTests();
    }


    @FXML
    private void onRestartAction() {
        connectivityTester.restartTests();
    }

    @FXML
    private void onSaveAction() {
        FileChooser fc = new FileChooser();
        fc.setTitle("Save Connectivity Test Result");
        fc.getExtensionFilters().add(new FileChooser.ExtensionFilter("Text file (*." + TESTRESULT_EXTENSION + ")", "*" + TESTRESULT_EXTENSION));
        fc.setInitialFileName("connectivity-test-" + FILE_DATE_TIME_FORMATTER.format(connectivityTester.getTestsStartDate()) + ".txt");


        //set the default save dir
        if (System.getProperty("user.home") != null)
            fc.setInitialDirectory(new File(System.getProperty("user.home")));

        if (jFedPreferences.containsPreference(GuiPreferenceKey.PREF_RSPECFOLDER_REQUEST)) {
            File dir = new File(jFedPreferences.getString(GuiPreferenceKey.PREF_RSPECFOLDER_REQUEST));
            if (dir.exists()) {
                fc.setInitialDirectory(dir);
            }
        }

        File file = fc.showSaveDialog(this.getScene().getWindow());

        if (file == null) //user cancelled the dialog
            return;


        //save folder in which the file was saved for future file selector dialogs
        jFedPreferences.setString(GuiPreferenceKey.PREF_RSPECFOLDER_REQUEST,
                file.getParentFile().getAbsolutePath());

        if (!file.getAbsolutePath().endsWith(TESTRESULT_EXTENSION))
            file = new File(file.getAbsolutePath() + TESTRESULT_EXTENSION);

        try {
            connectivityTester.saveResultsToFile(file);

            JFDialogs.create()
                    .owner(saveButton.getScene().getWindow())
                    .message(String.format("The connectivity test results were saved to '%s'", file.toString()))
                    .masthead("Successfully saved connectivity test results")
                    .title(null)
                    .showInformation();

        } catch (IOException e) {
            JFDialogs.create()
                    .owner(saveButton.getScene().getWindow())
                    .message("An error occurred while saving the connectivity test: " + e.getMessage())
                    .masthead("Saving connectivity test results failed")
                    .title("Saving connectivity test results failed")
                    .showException(e);
        }
    }
}
