package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.call_log_output.HtmlLogOutput;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.SfaExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartControllerManager;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.NonSfaExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.report.JobReportWrappingLogger;
import be.iminds.ilabt.jfed.highlevel.jobs.report.StitchingJobReport;
import be.iminds.ilabt.jfed.highlevel.jobs.sfa_parts.UpdateSfaExperimentPartUntilReadyStateSlice;
import be.iminds.ilabt.jfed.highlevel.jobs.states.*;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.tasks.GetAdvertisementTask;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.SCSRequestTask;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractGeniAggregateManager;
import be.iminds.ilabt.jfed.lowlevel.api.StitchingComputationService;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.stitching.StitchingDirector;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecLink;
import be.iminds.ilabt.jfed.rspec.model.StringRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.TargetAuthority;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.util.*;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 12/2/2015
 * Time: 16:43
 */
public class StartExperimentJob extends AbstractJob<Boolean> {
    private static final Logger LOG = LoggerFactory.getLogger(StartExperimentJob.class);

    private static final org.slf4j.Logger ACTUAL_LOG = LoggerFactory.getLogger(StartExperimentJob.class);
    private static final int TIME_BETWEEN_ALLOCATE_RETRIES_MS = 10_000;
    private static final int MAX_ALLOCATE_STITCH_ACTION_TRIES = 3;

    //this variable is typically set for debugging purposes
    private Integer waitForReadyDelay = null;

    private final SfaAuthority scsAuthority;
    private final JobStateFactory jobStateFactory;

    private final AuthorityListModel authorityListModel;
    private final AuthorityFinder authorityFinder;
    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    private final ExperimentPartControllerManager epcManager;

    private final StitchingJobReport stitchingJobReport;
    private final JobReportWrappingLogger log;
    private final JobReportWrappingLogger stitchingDirectorLog;

    private final AllocationDirector allocationDirector;


    private List<UserSpec> userSpecs = null;
    private Boolean success = null;
    private Collection<GeniUrn> failedResources = null; //TODO: register resources that could not be reserved here

    public StartExperimentJob(Experiment experiment, SfaAuthority scsAuthority, HighLevelTaskFactory hltf,
                              TaskThread tt, JobStateFactory jobStateFactory,
                              AuthorityListModel authorityListModel, AuthorityFinder authorityFinder,
                              AggregateManagerWrapperFactory aggregateManagerWrapperFactory, ExperimentPartControllerManager epcManager) {
        super("Start Experiment " + experiment.getName(), experiment, hltf, tt);
        this.scsAuthority = scsAuthority;
        this.jobStateFactory = jobStateFactory;
        this.authorityListModel = authorityListModel;
        this.authorityFinder = authorityFinder;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.epcManager = epcManager;

        assert experiment.getRequestedEndTime() != null;


        //do some initial setup for advanced logging

        this.stitchingJobReport = new StitchingJobReport(this);
        this.log = new JobReportWrappingLogger(ACTUAL_LOG, stitchingJobReport);
        this.stitchingDirectorLog = new JobReportWrappingLogger(StitchingDirector.ACTUAL_LOG, stitchingJobReport);

        if (experiment.getNewRequestRspecSource().getStringRspec().isStitching(authorityListModel, authorityFinder)) {
            this.allocationDirector =
                    new StitchingAllocationDirector(experiment, authorityListModel, authorityFinder, stitchingDirectorLog);

        } else {
            this.allocationDirector = new BasicAllocationDirector(experiment);
        }
    }

    @Override
    public Boolean execute() throws Exception {
        //-2. Do some preparations
        CreateUserSpecListState createUserSpecListState = jobStateFactory.createCreateUserSpecListState(experiment);
        setAndRunState(createUserSpecListState);

        userSpecs = createUserSpecListState.getUserSpecs();

        log.debug("StartStitchedExperimentJob started");

        //make sure we have a valid experiment
        assert experiment.getSlice() != null;
        assert experiment.getRequestedEndTime() != null;

        if (allocationDirector instanceof StitchingAllocationDirector) {

            SCSRequestState scsRequestState = new SCSRequestState();
            setAndRunState(scsRequestState);

            if (scsRequestState.getStatus() != ExperimentTaskStatus.SUCCESS) {
                log.error("SCS call failed", scsRequestState.getTaskExecution().getException());
                success = false;
                return false; //we give up
            }

            StitchingComputationService.ComputePathResult result = scsRequestState.getTask().getResult();
            assert result != null;

            //Update the slice request rspec with the enriched request-rspec from the SCS
            experiment.getSlice().setRequestRspec(new RequestRspecSource(new StringRspec(result.getServiceRspec()), ModelRspecType.FX));

            ((StitchingAllocationDirector) allocationDirector).setComputePathResult(result);
        }

        setAndRunState(new PopulateExperimentPartsState());

        if (hasNonSfaParts()) {
            setAndRunState(new CreateNonSfaExperimentPartsContainersState());

            setAndRunState(new StartNonSfaPartsState());
        }

        if (!hasSfaParts())
            return true;

        boolean permanentFailure = false;
        boolean temporaryFailure = false;

        do {
            temporaryFailure = false;

            final AllocateSfaPartsState allocateSfaPartsState = new AllocateSfaPartsState();
            setAndRunState(allocateSfaPartsState);

            if (allocateSfaPartsState.isPermanentFailure()) {
                permanentFailure = true;
                log.info("AllocatePartsState reported permanent failure");
                continue;
            }

            final ProvisionPartsState provisionPartsState = new ProvisionPartsState();
            setAndRunState(provisionPartsState);
            if (provisionPartsState.isPermanentFailure()) {
                permanentFailure = true;
                log.info("ProvisionPartsState reported permanent failure");
                continue;
            } else if (provisionPartsState.isTemporaryFailure()) {
                temporaryFailure = true;
                log.info("ProvisionPartsState reported temporary failure");
                continue;
            }

            final UpdateSfaPartsUntilReadyState updateUntilReadyState =
                    new UpdateSfaPartsUntilReadyState();
            setAndRunState(updateUntilReadyState);

            if (updateUntilReadyState.isPermanentFailure()) {
                permanentFailure = true;
                log.info("WaitForPartsReadyState reported permanent failure");
            } else if (updateUntilReadyState.isTemporaryFailure()) {
                temporaryFailure = true;
                log.info("WaitForPartsReadyState reported temporary failure");
            }
        } while (!permanentFailure && temporaryFailure);

        if (permanentFailure) {
            setAndRunState(jobStateFactory.createStopExperimentState(StartExperimentJob.this));

            success = false;
            return false;
        }

        //everything is now successfully running
        if (experiment.getSlice() != null && experiment.getSlice().getRequestRspec() != null) {
            boolean hasAutoShareLans = false;
            Collection<? extends RspecLink> links = experiment.getSlice().getRequestRspec().getModelRspec().getLinks();
            for (RspecLink link : links) {
                List<String> autoShareLanNames = link.getAutoShareLanNames();
                if (autoShareLanNames != null && !autoShareLanNames.isEmpty()) {
                    hasAutoShareLans = true;
                    break;
                }
            }
            if (hasAutoShareLans) {
                LOG.info("Experiment has auto shared LANs. Will share them.");
                final AutoShareLanPartsState autoShareLanPartsState = new AutoShareLanPartsState();
                setAndRunState(autoShareLanPartsState);

                LOG.debug("Experiment auto shared LANs have been shared. Will wait for ready again.");
                permanentFailure = false;
                do {
                    temporaryFailure = false;

                    final UpdateSfaPartsUntilReadyState updateUntilReadyState =
                            new UpdateSfaPartsUntilReadyState();
                    setAndRunState(updateUntilReadyState);

                    if (updateUntilReadyState.isPermanentFailure()) {
                        permanentFailure = true;
                        log.info("WaitForPartsReadyState reported permanent failure");
                    } else if (updateUntilReadyState.isTemporaryFailure()) {
                        temporaryFailure = true;
                        log.info("WaitForPartsReadyState reported temporary failure");
                    }
                } while (!permanentFailure && temporaryFailure);

                LOG.info("All is ready after sharing LANs -> all ready");
            } else
                LOG.info("Experiment has NO auto shared LANs -> all ready");
        } else
            LOG.warn("problem with experiment.getSlice().getRequestRspec()");

        success = true;
        return true;
    }

    public Boolean isSuccess() {
        return success;
    }

    public StitchingJobReport getStitchingJobReport() {
        return stitchingJobReport;
    }

    public Collection<GeniUrn> getFailedResources() {
        return failedResources;
    }

    public class SCSRequestState extends SingleCallState<SCSRequestTask> {
        //TODO: check mixed case of stitched and non-stitched authorities. Does SCS return non-stitched nodes/links in it's new RSpec?

        protected SCSRequestState() {
            super("Request stitched link setup-instructions",
                    hltf.createSCSRequestTask(experiment.getSlice(), experiment.getNewRequestRspecSource(),
                            experiment.getRequestedEndTime(), scsAuthority));
        }

        @Nonnull
        @Override
        public ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException {
            super.executeState(job);

            return getTask().getResult() != null ? ExperimentTaskStatus.SUCCESS : ExperimentTaskStatus.FAILED;
        }
    }

    private class PopulateExperimentPartsState extends State {
        protected PopulateExperimentPartsState() {
            super("Processing experiment information");
        }

        @Nonnull
        @Override
        public ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
            Set<TargetAuthority> authorities = new HashSet<>();

            //create SfaExperimentParts
            if (allocationDirector instanceof StitchingAllocationDirector) {
                Collection<SfaAuthority> auths = ((StitchingAllocationDirector) allocationDirector).getAllAuthorities();
                List<TargetAuthority> targetAuthorities = new ArrayList<>();
                //this is suboptimal, BUT it will work, as the central broker architecture (exogeni) requires
                // the central broker auth to be used for stitching
                for (SfaAuthority auth: auths) {
                    if (auth.hasSpecialCase(SfaAuthority.SpecialCase.CENTRAL_BROKER_SUBAUTH)) {
                        SfaAuthority connectAuth = authorityListModel.getByUrn(
                                auth.getUrn(),
                                AuthorityListModel.SubAuthMatchAllowed.ALLOW_TOPLEVEL,
                                AuthorityListModel.SubAuthMatchPreference.PREFER_TOPLEVEL,
                                true/*ignoreUrnTypeAndName*/);
                        if (connectAuth == null) {
                            log.error("ERROR: Component Manager Urn for central broker architecture toplevel authority in Request RSpec unknown (connectAuth=null): " +
                                    "\"" + auth.getUrn() + "\".");
                            throw new UnknownAuthorityException(auth.getUrn());
                        }
                        authorities.add(new TargetAuthority(auth.getUrn(), auth, connectAuth));
                    } else {
                        authorities.add(new TargetAuthority(auth.getUrn(), auth, auth));
                    }
                }
            } else {
                assert experiment.getSlice() != null;
                assert experiment.getSlice().getRequestRspec() != null;
                for (final GeniUrn componentManagerUrn : experiment.getSlice().getRequestRspec().getAllComponentManagerUrns()) {
                    final SfaAuthority connectAuth =
                            authorityFinder.findByUrn(componentManagerUrn, AuthorityFinder.Purpose.CREATE_SLIVER);
                    final SfaAuthority rspecAuth =
                            authorityFinder.findByUrn(componentManagerUrn, AuthorityFinder.Purpose.REQUEST_RSPEC);

                    if (connectAuth == null) {
                        log.error("ERROR: Component Manager Urn in Request RSpec unknown (connectAuth=null): \"" + componentManagerUrn + "\".");
                        throw new UnknownAuthorityException(componentManagerUrn);
                    }
                    if (rspecAuth == null) {
                        log.error("ERROR: Component Manager Urn in Request RSpec unknown (rspecAuth=null): \"" + componentManagerUrn + "\".");
                        throw new UnknownAuthorityException(componentManagerUrn);
                    }

                    if (connectAuth.isFake())
                        continue;

                    authorities.add(new TargetAuthority(componentManagerUrn, rspecAuth, connectAuth));
                }
            }

            for (TargetAuthority auth : authorities) {
                experiment.addExperimentPart(new SfaExperimentPart(experiment, auth.getSfaAuthorityToConnect(), auth.getSfaAuthorityForRspec()));
            }

            epcManager.populateNonSfaExperimentParts(experiment);


            return ExperimentTaskStatus.SUCCESS;
        }

    }

    public class AllocateSfaPartsState extends SlicedState<AllocatePartStateSlice> {
        private final List<AllocatePartStateSlice> slices = new ArrayList<>();
        private final Map<ExperimentPart, StateSliceThread<AllocatePartStateSlice>> threadMap = new HashMap<>();

        protected AllocateSfaPartsState() {
            super("Allocate resources with StitchingDirector");

            for (SfaExperimentPart experimentPart : getSfaParts()) {
                AllocatePartStateSlice allocatePartStateSlice = new AllocatePartStateSlice(experimentPart);
                slices.add(allocatePartStateSlice);
                StateSliceThread<AllocatePartStateSlice> thread =
                        new StateSliceThread<>(allocatePartStateSlice, getName());
                threadMap.put(experimentPart, thread);
            }

        }

        @Nonnull
        @Override
        protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
            threadMap.values().forEach(Thread::start);

            while (allocationDirector.areAnyActionsLeft()) {
                List<StitchingDirector.StitchAction> stitchActions = allocationDirector.getStitchActions();

                if (!stitchActions.isEmpty()) {
                    for (StitchingDirector.StitchAction stitchAction : stitchActions) {
                        ExperimentPart part = experiment.getPart(stitchAction.getAuthority());
                        threadMap.get(part).getStateSlice().submitStitchAction(new AllocateAction(stitchAction));
                    }
                } else {
                    log.trace("No stitchActions available for the moment, but director says it's not finished yet.");
                }

                Thread.sleep(250);
            }
            threadMap.values().forEach(apt -> apt.getStateSlice().submitStitchAction(AllocateAction.createFinishedAction()));

            return allocationDirector.isGaveUp() ? ExperimentTaskStatus.FAILED : ExperimentTaskStatus.SUCCESS;
        }

        public boolean isPermanentFailure() {
            return allocationDirector.isGaveUp();
        }

        public Map<ExperimentPart, StateSliceThread<AllocatePartStateSlice>> getThreadMap() {
            return threadMap;
        }

        public Collection<AllocatePartStateSlice> getSlices() {
            return slices;
        }
    }

    private static class AllocateAction {
        private final boolean finished;

        private final StitchingDirector.StitchAction stitchAction;

        public AllocateAction(StitchingDirector.StitchAction stitchAction) {
            this.stitchAction = stitchAction;
            this.finished = false;
        }

        private AllocateAction() {
            this.stitchAction = null;
            this.finished = true;
        }

        public static AllocateAction createFinishedAction() {
            return new AllocateAction();
        }

        public boolean isFinished() {
            return finished;
        }

        public StitchingDirector.StitchAction getStitchAction() {
            return stitchAction;
        }

        @Override
        public String toString() {
            return "AllocateAction{" +
                    "finished=" + finished +
                    ", stitchAction=" + stitchAction +
                    '}';
        }
    }

    public class AllocatePartStateSlice extends ExperimentPartStateSlice<SfaExperimentPart> {
        private final Logger log = StartExperimentJob.this.log;  //TODO: split into separate logger
        private final BlockingQueue<AllocateAction> queue = new LinkedBlockingQueue<>();

        private int allocateStitchActionRetryCount = 0;

        private AllocatePartStateSlice(SfaExperimentPart experimentPart) {
            super(StartExperimentJob.this, experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            while (true) {
                WaitForNextAllocateActionState waitForNextAllocateActionState = new WaitForNextAllocateActionState();
                setAndRunState(waitForNextAllocateActionState);

                AllocateAction allocateAction = waitForNextAllocateActionState.getAllocateAction();

                if (allocateAction.isFinished())
                    return ExperimentTaskStatus.SUCCESS;

                StitchingDirector.StitchAction stitchAction = allocateAction.getStitchAction();

                switch (stitchAction.getType()) {
                    case ALLOCATE:
                        processAllocateStitchAction(stitchAction);
                        break;
                    case DELETE:
                        processDeleteStitchAction();
                        break;
                    case LISTRESOURCES:
                        processListResourcesStitchAction();

                        break;
                    default:
                        throw new IllegalArgumentException("Unexpected StitchAction: " + stitchAction);
                }
            }

        }

        private void processListResourcesStitchAction() throws InterruptedException {
            GetAdvertisementState getAdvertisementState = new GetAdvertisementState();
            try {
                setAndRunState(getAdvertisementState);
            } catch (JFedException e) {
                log.error("Error while performing ListResources on {}", experimentPart.getConnectSfaAuthority(), e);
            }
            allocationDirector.processListResourcesResult(experimentPart.getConnectSfaAuthority(),
                    getAdvertisementState.getTask().getAdvertisementRspecString());
        }

        private void processDeleteStitchAction() throws InterruptedException {
            StopExperimentAtAuthorityPartState stopExperimentAtAuthorityPartState =
                    jobStateFactory.createStopExperimentAtAuthorityPartState(StartExperimentJob.this, experimentPart);
            try {
                setAndRunState(stopExperimentAtAuthorityPartState);
            } catch (JFedException e) {
                log.error("Error while deleting resources from experiment", e);
            } finally {
                allocationDirector.processDeleteResult(experimentPart.getConnectSfaAuthority());
            }
        }

        private void processAllocateStitchAction(StitchingDirector.StitchAction stitchAction) throws InterruptedException {
            RequestRspecSource overridingRequestRspecSource = stitchAction.getRequestRspec() != null ?
                    new RequestRspecSource(new StringRspec(stitchAction.getRequestRspec())) : null;

            final AllocateSliversOnAuthorityState allocateSliversOnAuthorityState =
                    jobStateFactory.createAllocateSliversOnAuthorityState(StartExperimentJob.this, experimentPart, userSpecs, overridingRequestRspecSource);


            AbstractGeniAggregateManager.AggregateManagerReply reply = null;

            try {
                setAndRunState(allocateSliversOnAuthorityState);

                if (allocateSliversOnAuthorityState.isSuccess()) {
                    reply = allocateSliversOnAuthorityState.getRawReply();

                } else if (allocateSliversOnAuthorityState.getException() instanceof AggregateManagerWrapper.VlanUnavailableException ||
                        allocateSliversOnAuthorityState.getException() instanceof AggregateManagerWrapper.AnyVlanUnsupportedException) {
                    JFedException e = (JFedException) allocateSliversOnAuthorityState.getException();
                    assert e.getSfaReply() == null || e.getSfaReply() instanceof AbstractGeniAggregateManager.AggregateManagerReply :
                            "Unexpected reply type: " + e.getSfaReply().getClass().getName();
                    //catch error, and report it to the director instead
                    reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
                } else if (allocateSliversOnAuthorityState.getException() instanceof JFedException) {

                    JFedException e = (JFedException) allocateSliversOnAuthorityState.getException();

                    if (e.getSfaReply() != null && e.getSfaReply() instanceof AbstractGeniAggregateManager.AggregateManagerReply) {
                        reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
                    } else {
                        if (StitchingDirector.isUnrecoverableError(e) == Boolean.TRUE) {
                            log.error("Giving up as {} reported an unrecoverable error {}", experimentPart.getName(), e.getMessage());
                            allocationDirector.reportGiveUp();
                        } else {
                            retryAllocate(stitchAction);
                            return; //we don't want to process this allocate-try any further
                        }
                    }
                } else {
                    log.error("Unexpected state: unknown error while allocating", allocateSliversOnAuthorityState.getException());
                }
            } catch (JFedException ex) {
                log.error("UNEXPECTED error while allocating", ex);
            }

            log.trace("   Processing result of CreateSliver call for " + experimentPart.getName());
            StitchingDirector.AllocateVerdict verdict = allocationDirector.processAllocateResult(stitchAction, reply);

            if (verdict == StitchingDirector.AllocateVerdict.CALLER_DELAYED_RETRY) {
                retryAllocate(stitchAction);
            }
        }

        private void retryAllocate(StitchingDirector.StitchAction stitchAction) throws InterruptedException {
            if (allocateStitchActionRetryCount++ > MAX_ALLOCATE_STITCH_ACTION_TRIES) {
                log.error("Giving up stitching as number of allocation tries on {} exceeded maximum of {}", experimentPart.getName(), MAX_ALLOCATE_STITCH_ACTION_TRIES);
                allocationDirector.reportGiveUp();
                return;
            }

            log.info("retryAllocate: CreateSliver failed (allocateStitchActionRetryCount=" + allocateStitchActionRetryCount + "), " +
                    "will delete and try again in " + TIME_BETWEEN_ALLOCATE_RETRIES_MS + "ms " +
                    "now=" + HtmlLogOutput.dateToString(new Date()));

            Thread.sleep(TIME_BETWEEN_ALLOCATE_RETRIES_MS);
            allocationDirector.reportReadyForRetry(experimentPart.getConnectSfaAuthority());
        }


        public void submitStitchAction(AllocateAction stitchAction) {
            boolean success = queue.offer(stitchAction);
            assert success : "Could not add StitchAction to ExperimentPart " + experimentPart.getName();
        }


        public class GetAdvertisementState extends SingleCallState<GetAdvertisementTask> {
            protected GetAdvertisementState() {
                super("Fetch advertisement from " + experimentPart.getName() + " for list of available VLAN's",
                        hltf.getAvailableAdverisement(experimentPart.getConnectSfaAuthority()));
            }
        }

        public class WaitForNextAllocateActionState extends be.iminds.ilabt.jfed.highlevel.jobs.State {

            private AllocateAction allocateAction;

            protected WaitForNextAllocateActionState() {
                super("Waiting for next StitchAction instruction");
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                allocateAction = queue.take();

                return ExperimentTaskStatus.SUCCESS;
            }

            public AllocateAction getAllocateAction() {
                return allocateAction;
            }
        }
    }

    public class AutoShareLanPartsState extends SlicedState<AutoShareLanPartJobSlice> {
        private final List<AutoShareLanPartJobSlice> slices;

        protected AutoShareLanPartsState() {
            super("Auto Sharing LANs in all ExperimentParts");

            slices = new ArrayList<>();

            List<SfaExperimentPart> sfaExperimentParts =
                    experiment.getParts().stream()
                            .filter(part -> part.getState() != InternalState.FAILED)
                            .filter(part -> part instanceof SfaExperimentPart)
                            .map(part -> (SfaExperimentPart) part)
                            .collect(Collectors.toList());

            Collection<? extends RspecLink> links = experiment.getSlice().getRequestRspec().getModelRspec().getLinks();

            for (SfaExperimentPart sfaExperimentPart : sfaExperimentParts) {
                SfaAuthority auth = sfaExperimentPart.getConnectSfaAuthority();
                boolean hasAutoShareLans = false;

                for (RspecLink link : links) {
                    List<String> autoShareLanNames = link.getAutoShareLanNames();
                    if (autoShareLanNames != null && !autoShareLanNames.isEmpty() && link.getComponentManagerUrns().contains(auth.getUrn())) {
                        hasAutoShareLans = true;
                        break;
                    }
                }

                if (hasAutoShareLans) {
                    LOG.debug("Auto Sharing LAN(s) in "+auth.getName());
                    AutoShareLanPartJobSlice slice = new AutoShareLanPartJobSlice(sfaExperimentPart);
                    slices.add(slice);
                } else {
                    LOG.debug("No Auto Shared LAN  in "+auth.getName());
                }
            }

            LOG.debug("AutoShareLanPartsState constructor finished with "+slices.size()+" slices.");

            assert !slices.isEmpty();
        }

        @Override
        public Collection<AutoShareLanPartJobSlice> getSlices() {
            return slices;
        }
    }

    public class ProvisionPartsState extends SlicedState<ProvisionSfaExperimentPartJobSlice> {
        private final List<ProvisionSfaExperimentPartJobSlice> slices;

        protected ProvisionPartsState() {
            super("Provision all ExperimentParts");

            slices =  experiment.getParts().stream()
                    .filter(part -> part.getState() != InternalState.FAILED)
                    .filter(part -> part instanceof SfaExperimentPart)
                    .map(part -> new ProvisionSfaExperimentPartJobSlice((SfaExperimentPart) part))
                    .collect(Collectors.toList())   ;
        }


        public boolean isPermanentFailure() {
            return false;
        }

        public boolean isTemporaryFailure() {
            return false;
        }

        @Override
        public Collection<ProvisionSfaExperimentPartJobSlice> getSlices() {
            return slices;
        }
    }

    public class ProvisionSfaExperimentPartJobSlice extends ExperimentPartStateSlice<SfaExperimentPart> {

        public ProvisionSfaExperimentPartJobSlice(SfaExperimentPart experimentPart) {
            super(StartExperimentJob.this, experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            if (!supportsAllocateProvision(experimentPart.getConnectSfaAuthority())) {
                log.debug("Skipping provisioning on {}, as it does not support it.", experimentPart.getName());
                return ExperimentTaskStatus.SUCCESS;
            }

            ProvisionSliversOnAuthorityState provisionSliversOnAuthorityState =
                    jobStateFactory.createProvisionSliversOnAuthorityState(StartExperimentJob.this, experimentPart, userSpecs);

            setAndRunState(provisionSliversOnAuthorityState);

            return provisionSliversOnAuthorityState.getStatus();
        }

        private boolean supportsAllocateProvision(SfaAuthority auth) {
            AggregateManagerWrapper amWrapper =
                    aggregateManagerWrapperFactory.getAggregateManagerWrapper(auth);

            return amWrapper.hasSeperateAllocateAndProvision();
        }
    }

    public class AutoShareLanPartJobSlice extends ExperimentPartStateSlice<SfaExperimentPart> {

        public AutoShareLanPartJobSlice(SfaExperimentPart experimentPart) {
            super(StartExperimentJob.this, experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            if (!supportsShareLan(experimentPart.getConnectSfaAuthority())) {
                log.warn("Share LAN on {} is not supported. Will try anyway.", experimentPart.getName());
                //return ExperimentTaskStatus.WARNING;
            }
            Collection<? extends RspecLink> links = experiment.getSlice().getRequestRspec().getModelRspec().getLinks();

            ExperimentTaskStatus worstStatus = null;
            SfaAuthority auth = experimentPart.getConnectSfaAuthority();
            LOG.debug("Start sharing LAN(s) on "+auth.getHrn());
            for (RspecLink link : links) {
                List<String> autoShareLanNames = link.getAutoShareLanNames();
                if (autoShareLanNames != null && !autoShareLanNames.isEmpty() && link.getComponentManagerUrns().contains(auth.getUrn())) {
                    String sharedLanName = autoShareLanNames.get(0);

                    if (autoShareLanNames.size() > 1)
                        LOG.warn("There is more than one auto share LAN name per link (for link\""+link.getClientId()+"\"), ignoring all but first.");

                    LOG.debug("Sharing LAN "+link.getClientId()+" as "+sharedLanName+" on "+auth.getHrn());

                    ShareLanAtAuthorityPartState shareLanState =
                            jobStateFactory.createShareLanState(StartExperimentJob.this, experimentPart, link.getClientId(), sharedLanName);

                    setAndRunState(shareLanState);
                    ExperimentTaskStatus status = shareLanState.getStatus();
                    LOG.debug("Shared LAN "+link.getClientId()+" as "+sharedLanName+" on "+auth.getHrn()+" -> status = "+status);

                    if (status != ExperimentTaskStatus.SUCCESS || worstStatus == null) {
                        worstStatus = status;
                    }
                }
            }

            //there should be at least one link shared, or this shouldn't have been called.
            assert worstStatus != null;

            LOG.debug("Stopped sharing LAN(s) on "+auth.getHrn()+" -> worst status = "+worstStatus);

            return worstStatus;
        }

        private boolean supportsShareLan(SfaAuthority auth) {
            AggregateManagerWrapper amWrapper =
                    aggregateManagerWrapperFactory.getAggregateManagerWrapper(auth);

            return amWrapper.isShareLanSupported();
        }
    }

    public class UpdateSfaPartsUntilReadyState extends SlicedState<UpdateSfaExperimentPartUntilReadyStateSlice> {
        private final List<UpdateSfaExperimentPartUntilReadyStateSlice> slices;

        private UpdateSfaPartsUntilReadyState() {
            super("Waiting for resources to become ready");

            slices = experiment.getParts().stream()
                    .filter(part -> part.getState() != InternalState.FAILED)
                    .filter(part -> part instanceof SfaExperimentPart)
                    .map(part -> new UpdateSfaExperimentPartUntilReadyStateSlice(StartExperimentJob.this, (SfaExperimentPart) part, waitForReadyDelay, jobStateFactory))
                    .collect(Collectors.toList());
        }


        public boolean isPermanentFailure() {
            return getSlices().stream().anyMatch(updateSlice -> updateSlice.getResult() != ExperimentTaskStatus.SUCCESS);
        }

        public boolean isTemporaryFailure() {
            return false;
        }

        @Override
        public Collection<UpdateSfaExperimentPartUntilReadyStateSlice> getSlices() {
            return slices;
        }
    }

    public void setWaitForReadyDelay(Integer waitForReadyDelay) {
        this.waitForReadyDelay = waitForReadyDelay;
    }


    private Collection<SfaExperimentPart> getSfaParts() {
        return experiment.getParts().stream()
                .filter(part -> part instanceof SfaExperimentPart)
                .map(SfaExperimentPart.class::cast)
                .collect(Collectors.toList());

    }


    private Collection<NonSfaExperimentPartController> getNonSfaExperimentPartControllers() {

        Set<NonSfaExperimentPartController> controllers = new HashSet<>();

        for (ExperimentPart experimentPart : experiment.getParts()) {
            if (experimentPart instanceof SfaExperimentPart)
                continue;


            ExperimentPartController epc = epcManager.getController(experimentPart);
            if (!(epc instanceof NonSfaExperimentPartController))
                throw new RuntimeException("Expected a non-sfa ExperimentPartController");

            controllers.add((NonSfaExperimentPartController) epc);
        }

        return controllers;
    }

    private boolean hasNonSfaParts() {
        return experiment.getParts().stream().anyMatch(part -> !(part instanceof SfaExperimentPart));
    }

    private boolean hasSfaParts() {
        return experiment.getParts().stream().anyMatch(part -> (part instanceof SfaExperimentPart));
    }

    public final class CreateNonSfaExperimentPartsContainersState extends SlicedState<StateSlice> {

        private final List<StateSlice> slices;

        protected CreateNonSfaExperimentPartsContainersState() {
            super("Creating non-SFA experimentparts containers");
            slices = getNonSfaExperimentPartControllers().stream()
                    .map(epc -> epc.initializeContainer(StartExperimentJob.this))
                    .collect(Collectors.toList());
        }

        @Override
        public Collection<StateSlice> getSlices() {
            return slices;
        }
    }


    public final class StartNonSfaPartsState extends SlicedState<StateSlice> {

        private final List<StateSlice> slices;

        protected StartNonSfaPartsState() {
            super("Starting non-SFA resources");

            slices = new ArrayList<>();

            for (ExperimentPart experimentPart : experiment.getParts()) {
                if (experimentPart instanceof SfaExperimentPart)
                    continue;

                ExperimentPartController epc = epcManager.getController(experimentPart);
                if (!(epc instanceof NonSfaExperimentPartController))
                    throw new RuntimeException("Expected a non-sfa ExperimentPartController");

                //noinspection unchecked
                slices.add(
                        ((NonSfaExperimentPartController) epc).start(StartExperimentJob.this, experimentPart));

            }
        }

        @Override
        public Collection<StateSlice> getSlices() {
            return slices;
        }
    }

}
