package be.iminds.ilabt.jfed.lowlevel.userloginmodel;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.KeyUtil;
import org.apache.commons.codec.binary.Base64;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.*;
import java.security.KeyPair;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.cert.CertificateParsingException;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateCrtKey;
import java.security.interfaces.RSAPublicKey;
import java.util.Collection;
import java.util.List;

/**
 * This is a UserLoginModel that reads a key and certificate, and automatically extracts the authority and user URN.
 * <p>
 * It also decodes and stores the private key using the password
 */
public class KeyCertUserLoginModel implements UserLoginModel {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(KeyCertUserLoginModel.class);
    private static final String PREFS_KEY_CERT_FILE_URI = "keyCertFileURI";
    private static final String PREFS_KEY_SOURCE = "source";
    private static final String PREFS_KEY_FILE = "file";
    private static final String PREFS_KEY_CERT_FILE_CONTENT = "keyCertFileContent";

    protected File keyCertFile;
    protected URL keyCertUrl;
    protected String keyCertContent;
    protected List<X509Certificate> certificateChain;
    protected PrivateKey privateKey; //unencrypted, so never saved directly!
    protected RSAPublicKey publicKey;
    protected SfaAuthority authority;
    protected String authorityUrn;
    protected GeniUrn userUrn;
    protected boolean passwordRequired;
    protected AuthorityListModel authorityListModel;
    protected AuthorityFinder authorityFinder;
    protected UserLoginModelManager userLoginModelManager;
    private PemSource pemSource;

    public KeyCertUserLoginModel(AuthorityListModel authorityListModel, AuthorityFinder authorityFinder, UserLoginModelManager userLoginModelManager) {
        this.authorityListModel = authorityListModel;
        this.authorityFinder = authorityFinder;
        this.userLoginModelManager = userLoginModelManager;

        this.keyCertFile = null;
        this.pemSource = PemSource.FILE;
        this.keyCertContent = null;
        this.authority = null;
        this.authorityUrn = null;
        this.userUrn = null;
    }

    /**
     * Override current values with keyCertFile and automatic values derived from keyCertFile
     * <p>
     *
     * @throws be.iminds.ilabt.jfed.lowlevel.userloginmodel.InvalidLoginException
     */
    public void setKeyCertPemFile(File newKeyCertFile) throws InvalidLoginException {
        reset();

        if (newKeyCertFile == null) {
            return;
        }

        if (!newKeyCertFile.exists())
            throw new InvalidLoginException("Key and Certificate file does not exist: \"" + newKeyCertFile.getPath() + "\"\n");

        try {
            this.keyCertFile = newKeyCertFile;
            this.keyCertContent = IOUtils.fileToString(newKeyCertFile);

            if (LOG.isDebugEnabled()) {
                String dataMd5String = keyCertContent == null ? "null" : "empty";
                if (keyCertContent != null && !keyCertContent.isEmpty()) {
                    try {
                        MessageDigest md = MessageDigest.getInstance("MD5");
                        byte[] dataMd5 = md.digest(keyCertContent.getBytes("UTF-8"));
                        dataMd5String = Base64.encodeBase64String(dataMd5);
                    } catch (NoSuchAlgorithmException e) {
                        LOG.debug("Error creating read PEM MD5", e);
                        dataMd5String = "error";
                    }
                }

                LOG.debug("Read RSA Private key and certificate data from PEM file \"" + newKeyCertFile.getAbsolutePath() + "\"." +
                        " len=" + (keyCertContent == null ? "null" : keyCertContent.length()) +
                        " md5="+dataMd5String);
            }
        } catch (IOException e) {
            throw new InvalidLoginException("Error reading \"" + keyCertFile.getPath() + "\": " + e.getMessage() + "\n", e);
        }

        assert this.keyCertFile != null;
        processPemContent();
    }

    public void setKeyCertPemUrl(URL url) throws InvalidLoginException {
        assert url != null;

        reset();
        this.pemSource = PemSource.URL;
        this.keyCertUrl = url;
        try {
            URLConnection conn = url.openConnection();
            BufferedReader br = new BufferedReader(new InputStreamReader(conn.getInputStream()));

            StringBuilder sb = new StringBuilder();
            try {

                String line;
                while ((line = br.readLine()) != null) {
                    sb.append(line);
                    sb.append("\n");
                }
            } finally {
                br.close();
            }

            String newKeyCertContent = sb.toString();

            if (newKeyCertContent.isEmpty()) {
                throw new InvalidLoginException("Could not read key-certificate pair from the provided URL. Got null");
            }

            this.keyCertContent = newKeyCertContent;

            processPemContent();

        } catch (IOException e) {
            LOG.error("Could not fetch key-certificate-pair", e);
            throw new InvalidLoginException("Could not fetch key-certificate-pair", e);
        }

    }


    /**
     * Override current values with String contents and automatic values derived from keyCertFile
     * <p>
     *
     * @throws be.iminds.ilabt.jfed.lowlevel.userloginmodel.InvalidLoginException
     */
    public void setKeyCertPemString(String newKeyCertContent) throws InvalidLoginException {
        assert newKeyCertContent != null;
        assert !newKeyCertContent.isEmpty();

        reset();
        this.pemSource = PemSource.STRING;
        this.keyCertContent = newKeyCertContent;
        processPemContent();

    }

    /**
     * Reads the information from the PEM-file/string
     *
     * @throws InvalidLoginException
     */
    protected void processBasicPemContent() throws InvalidLoginException {
        this.authority = null;
        this.authorityUrn = null;
        this.userUrn = null;

        if (keyCertContent == null)
            throw new InvalidLoginException("Key and Certificate PEM is empty");

        if (!keyCertContent.trim().startsWith("-----BEGIN"))
            throw new InvalidLoginException("Key and Certificate PEM does not have expected content");

        certificateChain = KeyUtil.pemToX509CertificateChain(keyCertContent);
        if (certificateChain == null)
            throw new InvalidLoginException("Error parsing certificate PEM");

        boolean hasKey = KeyUtil.hasAnyPrivateKey(keyCertContent);
        if (hasKey) {
            boolean encryptedKey = KeyUtil.hasEncryptedRsaPrivateKey(keyCertContent);
            this.passwordRequired = encryptedKey;
            if (!encryptedKey)
                try {
                    this.privateKey = KeyUtil.pemToAnyPrivateKey(keyCertContent, null);
                } catch (KeyUtil.PEMDecodingException pemDecodingError) {
                    throw new InvalidLoginException("Error reading private key", pemDecodingError);
                }
        } else {
            throw new InvalidLoginException("No private key found in PEM");
        }
    }

    protected void processPemContent() throws InvalidLoginException {
        //these two methods are not unified into one method because KeyCertWithManualInfoUserLoginModel does not use deriveFromPemContent()
        processBasicPemContent();
        deriveFromPemContent();
    }

    /**
     * Derive the associated authority from the PEM-file/string
     *
     * @throws InvalidLoginException
     */
    protected void deriveFromPemContent() throws InvalidLoginException {
        assert certificateChain != null;
        assert !certificateChain.isEmpty();
        Collection<List<?>> altNames;
        try {
            X509Certificate firstCertificate = certificateChain.get(0);
            altNames = firstCertificate.getSubjectAlternativeNames();
        } catch (CertificateParsingException e) {
            throw new InvalidLoginException("Error processing certificate alternate names", e);
        }

        if (altNames != null) {
            for (List<?> altName : altNames) {
                Integer nameType = (Integer) altName.get(0);
                if (nameType == 6) {
                    //uniformResourceIdentifier
                    String urn = (String) altName.get(1);
                    GeniUrn geniUrn = GeniUrn.parse(urn);
                    if (geniUrn == null || !(geniUrn.getEncodedResourceType().equals("user")
                            || geniUrn.getEncodedResourceType().equals("tool"))) {
                        LOG.warn("Certificate alternative name URI is not a valid user or tool urn: '{}'" +
                                " (will be ignored)", urn);
                    } else {
                        userUrn = geniUrn;
                        authority = authorityFinder.findByAnyUrn(userUrn, AuthorityFinder.Purpose.FIND_USERAUTH);
                        if (authority != null)
                            break;
                        else
                            LOG.warn("User authority found '{}', but no authority info about it is known. " +
                                    "Try adding authority info to the internal list and try again.", geniUrn);
                    }

                }
            }
        } else
            LOG.warn("No alternative names in certificate");

        if (authority != null) {
            if (authorityUrn == null)
                authorityUrn = authority.getUrnString();
        } else {
            throw new InvalidLoginException("ERROR: Did not find info about user (urn and authority) in certificate.");
        }
    }

    public void save(UserLoginModelManager.UserLoginModelPreferences prefs) {
        LOG.trace("KeyCertUserLoginModel.save(prefs)");

        assert (pemSource != null);

        prefs.put(PREFS_KEY_SOURCE, pemSource.name());

        if (pemSource == PemSource.FILE) {
            prefs.remove(PREFS_KEY_CERT_FILE_CONTENT);
            if (keyCertFile != null)
                prefs.put(PREFS_KEY_CERT_FILE_URI, keyCertFile.toURI().toString());
            else
                prefs.remove(PREFS_KEY_CERT_FILE_URI);
        } else if (pemSource == PemSource.URL) {
            prefs.remove(PREFS_KEY_CERT_FILE_CONTENT);
            if (keyCertUrl != null)
                prefs.put(PREFS_KEY_CERT_FILE_URI, keyCertUrl.toExternalForm());
            else {
                prefs.remove(PREFS_KEY_CERT_FILE_URI);
            }
        } else {
            assert (pemSource == PemSource.STRING);
            prefs.remove(PREFS_KEY_CERT_FILE_URI);
            //don't save content of keycert in preferences
//            if (keyCertContent != null)
//                prefs.put(PREFS_KEY_CERT_FILE_CONTENT, keyCertContent);
//            else
            prefs.remove(PREFS_KEY_CERT_FILE_CONTENT);
        }
    }

    public void reset() {
        this.keyCertFile = null;
        this.keyCertUrl = null;
        this.pemSource = PemSource.FILE;
        this.keyCertContent = null;
        this.authority = null;
        this.authorityUrn = null;
        this.userUrn = null;
        this.privateKey = null;
        this.publicKey = null;
        this.certificateChain = null;
    }

    public void defaults() {
        reset();
    }

    public void load(UserLoginModelManager.UserLoginModelPreferences prefs) throws InvalidLoginException {
        reset();

        if (prefs.contains(PREFS_KEY_SOURCE)) {
            if (prefs.get(PREFS_KEY_SOURCE).equals(PemSource.FILE.name()))
                loadFromFile(prefs);
            else if (prefs.get(PREFS_KEY_SOURCE).equals(PemSource.URL.name()))
                loadFromUrl(prefs);
            else {
                assert prefs.get(PREFS_KEY_SOURCE).equals(PemSource.STRING.name());
//                loadFromString(prefs);
            }

        } else {
            //fallback to old login settings source
            boolean fromFile = prefs.getBoolean(PREFS_KEY_FILE, true);
            if (fromFile)
                loadFromFile(prefs);
            //         else
            //              loadFromString(prefs);
        }
    }

    private void loadFromFile(UserLoginModelManager.UserLoginModelPreferences prefs) throws InvalidLoginException {
        if (prefs.contains(PREFS_KEY_CERT_FILE_URI)) {
            String keyCertFileUri = prefs.get(PREFS_KEY_CERT_FILE_URI);
            File file;
            try {
                file = new File(new URI(keyCertFileUri));
            } catch (URISyntaxException e) {
                LOG.warn("Stored file URI is not a valid URI. It will be deleted from the settings", e);
                prefs.remove(PREFS_KEY_CERT_FILE_URI);
                return;
            }

            setKeyCertPemFile(file);

        }
    }

    // don't save content of certificate in preferences for security purposes
//    private void loadFromString(UserLoginModelManager.UserLoginModelPreferences prefs) throws InvalidLoginException {
//        String loadedKeyCertContent = prefs.get(PREFS_KEY_CERT_FILE_CONTENT, null);
//
//        if (loadedKeyCertContent != null)
//            setKeyCertPemString(loadedKeyCertContent);
//        else {
//            LOG.warn("Could not fetch String certificate content from preferences");
//        }
//    }

    private void loadFromUrl(UserLoginModelManager.UserLoginModelPreferences prefs) throws InvalidLoginException {
        if (prefs.contains(PREFS_KEY_CERT_FILE_URI)) {
            String keyCertFileUri = prefs.get(PREFS_KEY_CERT_FILE_URI);
            URL url;
            try {
                url = new URL(keyCertFileUri);
            } catch (MalformedURLException e) {
                LOG.warn("Stored URL is not a valid. It will be deleted from the settings", e);
                prefs.remove(PREFS_KEY_CERT_FILE_URI);
                return;
            }

            setKeyCertPemUrl(url);

        }
    }

    /**
     * password is not stored internally
     *
     * @return success
     */
    public boolean unlock(String password) {
        return unlock(password.toCharArray());
    }

    /**
     * password is not stored internally
     *
     * @return success
     */
    public boolean unlock(char[] password) {
        if (keyCertContent == null) return false;

        try {
            this.privateKey = KeyUtil.pemToAnyPrivateKey(keyCertContent, password);
            if (this.privateKey instanceof RSAPrivateCrtKey) {
                this.publicKey = KeyUtil.rsaPrivateKeyToRsaPublicKey((RSAPrivateCrtKey) this.privateKey);
                if (this.publicKey == null) {
                    LOG.warn("rsaPrivateKeyToRsaPublicKey failed! Converting private key to public key using fallback pemToRsaKeyPair");
                    //fallback
                    KeyPair pair = KeyUtil.pemToRsaKeyPair(keyCertContent, password);
                    if (pair == null) {
                        LOG.error("ERROR: decoded private RSA key, but failed to create key pair. privateKey class = " + privateKey.getClass().getName());
                        this.publicKey = null;
                        this.privateKey = null;
                    } else {
                        if (pair.getPublic() == null || !(pair.getPublic() instanceof RSAPublicKey)) {
                            LOG.error("ERROR: decoded private RSA key, but key pair public key is not RSAPublicKey: " +
                                    (pair == null ? null : pair.getClass().getName()) +
                                    "  privateKey class=" + privateKey.getClass().getName());
                            this.publicKey = null;
                            this.privateKey = null;
                        } else
                            this.publicKey = (RSAPublicKey) pair.getPublic();
                    }
                }
            }
            if (privateKey == null)
                LOG.trace("ERROR: Failed to decode private key using password");
        } catch (KeyUtil.PEMDecodingException pemDecodingError) {
            LOG.trace("ERROR: Failed to decode private key using password", pemDecodingError);
            this.privateKey = null;
        }
        return this.privateKey != null;
    }

    /**
     * Did the user specify a PEM file or String yet?
     */
    public boolean isPemGiven() {
        return keyCertContent != null || keyCertFile != null;
    }

    /**
     * Only valid if isPemGiven() is true!
     * Did the user provide a filename to read the private key and certificate from?
     */
    public boolean isFromFile() {
        return pemSource == PemSource.FILE;
    }

    /**
     * Only valid if isPemGiven() is true!
     * Did the user provide a String to read the private key and certificate from?
     */
    public boolean isFromString() {
        return pemSource == PemSource.STRING;
    }

    public boolean isFromURL() {
        return pemSource == PemSource.URL;
    }

    public boolean isPasswordRequired() {
        return passwordRequired;
    }

    public File getKeyCertFile() {
        return keyCertFile;
    }

    @Override
    public File getPrivateKeyFile() {
        return keyCertFile;
    }

    @Override
    public File getCertificateFile() {
        return keyCertFile;
    }

    public String getKeyCertContent() {
        return keyCertContent;
    }

    @Override
    public GeniUrn getUserUrn() {
        return userUrn;
    }

    @Override
    public String getUserUrnString() {
        return userUrn != null ? userUrn.getValue() : null;
    }

    @Override
    public PrivateKey getPrivateKey() {
        return privateKey;
    }

    @Override
    public List<X509Certificate> getClientCertificateChain() {
        return certificateChain;
    }

    @Override
    public SfaAuthority getUserAuthority() {
        return authority;
    }

    public String getUserAuthorityUrn() {
        return authorityUrn;
    }

    public boolean isValid() {
        if (authority == null) return false;
        if (userUrn == null) return false;
        if (keyCertContent == null) return false;
        if (privateKey == null) return false;
        if (certificateChain == null) return false;
        if (certificateChain.isEmpty()) return false;
        return true;
    }

    public RSAPublicKey getPublicKey() {
        return publicKey;
    }

    @Override
    public GeniUser getLoggedInGeniUser() {
        if (!isValid()) return null;
        return this;
    }

    @Override
    public boolean isUserLoggedIn() {
        return isValid();
    }

    @Override
    public boolean equals(Object o) {
        throw new UnsupportedOperationException();
    }

    @Override
    public int hashCode() {
        throw new UnsupportedOperationException();
    }

    public URL getKeyCertURL() {
        return keyCertUrl;
    }


    private static enum PemSource {FILE, URL, STRING}
}
